from dataclasses import dataclass, field
from pathlib import Path

from slugify import slugify

from . import iohandler, sources


@dataclass(frozen=True)
class BaseMetadata:
    title: str
    authors: list[str]
    url: str
    genres: list[str]
    description: str
    cover_art: str

    def __post_init__(self) -> None:
        self.title_slug = slugify(self.title)

    def asdict(self) -> dict:
        return {
            "title": self.title,
            "authors": self.authors,
            "url": self.url,
            "genres": self.genres,
            "description": self.description,
            "cover_art": self.cover_art,
        }


@dataclass
class BaseChapter:
    title: str
    url: str
    slug: str = None

    def __post_init__(self) -> None:
        if self.slug is None:
            self.slug = slugify(self.title)

    def asdict(self) -> dict:
        return {
            "title": self.title,
            "url": self.url,
            "slug": self.slug,
        }


@dataclass
class ComicChapter(BaseChapter):
    _images: list[str] = field(init=False, default_factory=list)

    @property
    def images(self) -> list[str]:
        if self._images:
            return self.images

        self._images = 0


@dataclass
class Comic:
    def __init__(
        self,
        metadata: BaseMetadata,
        chapters: list[BaseChapter],
    ):
        self.metadata = metadata
        self.chapters = chapters
        self.source = sources.get_class_for(self.metadata.url)(self.metadata.url)

    def asdict(self) -> dict:
        return {
            "metadata": self.metadata.asdict(),
            "chapters": [c.asdict() for c in self.chapters],
        }

    def get_chapter_image_urls(self, chapter: BaseChapter) -> list[str]:
        return self.source.fetch_chapter_image_list(chapter)

    def save_metadata(self, path: Path | str) -> None:
        iohandler.save_comic(self, path)



from pathlib import Path
from typing import Iterable

from ..comic import Comic
from . import ConvertOptions

ACCEPTED_IMAGE_EXTENSIONS = {
    ".gif": "gif",
    ".png": "png",
    ".jpg": "jpeg",
    ".jpeg": "jpeg",
    ".jpe": "jpeg",
}


class BaseConverter:
    def __init__(self, comic: Comic, options: ConvertOptions | None = None) -> None:
        self.comic = comic
        self.options = options or ConvertOptions()

    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterable[None]:
        raise NotImplementedError

    def create_file(self, path: Path | str, save_to: Path | str) -> None:
        for _ in self.create_file_progress(path, save_to):
            pass



# pylint: disable=no-member

import shutil
import tempfile
import zipfile
from pathlib import Path
from typing import Iterable

from lxml import etree
from lxml.builder import E, ElementMaker

from ..iohandler import NUM_LEFT_PAD_DIGITS
from .base_converter import ACCEPTED_IMAGE_EXTENSIONS, BaseConverter

STYLE_CSS = """
@page {
    margin: 0
}

body {
    display: block;
    margin: 0;
    padding: 0;
}
""".strip()

CONTAINER_XML = """
<?xml version="1.0"?>
<container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">
<rootfiles>
<rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>
</rootfiles>
</container>
""".strip()

OPF_MAP = {"opf": "http://www.idpf.org/2007/opf"}
HTML_NS = "http://www.w3.org/1999/xhtml"


class EpubConverter(BaseConverter):
    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterable[None]:
        path = Path(path)
        save_to = Path(save_to)
        with tempfile.TemporaryDirectory() as temp:
            root = Path(temp)
            oebps = self.create_skeleton(root)
            self.generate_content_opf()
            self.generate_nav_xhtml()
            self.generate_toc_ncx()

            for chap in self.comic.chapters:
                yield "Building"
                for i, image in enumerate((path / chap.slug).iterdir()):
                    if image.suffix in ACCEPTED_IMAGE_EXTENSIONS:
                        new_file = (
                            oebps
                            / "Text"
                            / chap.slug
                            / f"{i:{NUM_LEFT_PAD_DIGITS}}.xhtml"
                        )
                        new_file.write_text(self.generate_image_html(chap.slug, image))

                        shutil.copyfile(
                            image,
                            oebps
                            / "Images"
                            / chap.slug
                            / f"{i:{NUM_LEFT_PAD_DIGITS}}{image.suffix}",
                        )

            dest_file = save_to / f"{self.metadata.title}.epub"
            with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as zipout:
                zipout.writestr("mimetype", "application/epub+zip", zipfile.ZIP_STORED)
                for p in root.rglob("*"):
                    yield "Compressing"
                    if p.is_file():
                        relative_path = Path(str(p).lstrip(str(root)))
                        zipout.write(p, relative_path, zipfile.ZIP_DEFLATED)

    def create_skeleton(self, tmpdir: Path) -> Path:
        """
        Create the following structure at `tmpdir`:
        ```
        tmpdir/
          - META-INF/container.xml
          - OEBPS/
              - Images/
              - Text/
        ```
        :return The OEBPS directory.
        """
        meta_inf = tmpdir / "META-INF"
        meta_inf.mkdir()
        container_xml = meta_inf / "container.xml"
        container_xml.write_text(CONTAINER_XML, encoding="utf-8")

        oebps = tmpdir / "OEBPS"
        image_dir = oebps / "Images"
        text_dir = oebps / "Text"

        text_dir.mkdir(parents=True)
        image_dir.mkdir(parents=True)

        return oebps

    def generate_image_html(
        self,
        chapter_slug: str,
        image_path: Path,
    ) -> None:
        image_base = image_path.with_suffix("")
        tree = E.html(
            E.head(
                E.title(image_base),
                E.link(href="../style.css", type="text/css", rel="stylesheet"),
            ),
            E.body(
                E.div(style="text-align:center;top:0.0%"),
                E.img(
                    width="auto",
                    height="100%",
                    src=f"../../Images/{chapter_slug}/{image_path}",
                ),
            ),
        )

        return etree.tostring(
            tree,
            xml_declaration=True,
            encoding="utf-8",
            pretty_print=True,
        ).decode("utf-8")

    def generate_toc_ncx(self) -> str:
        nav_map: list[etree._Element] = []

        for chap in self.comic.chapters:
            nav_map.append(
                E.navPoint(
                    E.navLabel(
                        E.text(chap.title),
                    ),
                    E.content(src=f"Text/{chap.slug}/{1:{NUM_LEFT_PAD_DIGITS}}.xhtml"),
                    id=f"Text_{chap.slug}",
                )
            )

        tree = E.ncx(
            E.head(E.meta()),
            E.docTitle(E.text(self.comic.metadata.title)),
            E.navMap(*nav_map),
        )

        return etree.tostring(
            tree,
            xml_declaration=True,
            encoding="utf-8",
            pretty_print=True,
        ).decode("utf-8")

    def generate_content_opf(self, cover: Path | None = None) -> str:
        M = ElementMaker(  # pylint: disable=invalid-name
            namespace="http://purl.org/dc/elements/1.1/",
            nsmap=OPF_MAP,
        )

        metadata = self.comic.metadata
        els_genres = [M.subject(i) for i in metadata.genres]
        els_authors = [M.creator(i) for i in metadata.authors]

        if cover:
            cover_el = E.item(
                id="cover",
                href=f"Images/{cover.name}",
                properties="cover-image",
            )
            cover_el.attrib[
                "media-type"
            ] = f"image/{ACCEPTED_IMAGE_EXTENSIONS[cover.suffix]}"
        else:
            cover_el = None

        package = E.package(
            M.metadata(
                M.title(self.title),
                *els_authors,
                M.contributor("mandown", role="bkp"),
                M.description(metadata.description),
                M.identifier(metadata.url, scheme="URL"),
                M.language("eng"),
                *els_genres,
                M.meta(name="mandown:cover", content=self.cover_art),
                # TODO: fix cover, incorporate it in
                # TODO: also add in chapters and orientation and metas
                E.guide(E.reference(type="cover", title="Cover", href="cover")),
                nsmap=OPF_MAP["opf"],
                version="3.0",
            )
        )

        return etree.tostring(
            package,
            xml_declaration=True,
            encoding="utf-8",
            pretty_print=True,
        )

    def generate_nav_xhtml(self) -> str:
        pass


def get_class() -> EpubConverter:
    return EpubConverter



import os
import zipfile
from pathlib import Path
from typing import Iterable

from .base_converter import ACCEPTED_IMAGE_EXTENSIONS, BaseConverter


class CbzConverter(BaseConverter):
    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterable[None]:
        path = Path(path)
        save_to = Path(save_to)

        dest_file = save_to / f"{self.comic.metadata.title}.cbz"

        with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as file:
            for dirpath, _, filenames in os.walk(path):
                yield "Compressing"
                for name in filenames:
                    if (Path(dirpath) / name).is_file():
                        file.write(
                            Path(dirpath) / name,
                            Path(dirpath.lstrip(str(path))) / name,
                            zipfile.ZIP_DEFLATED,
                        )


def get_class() -> CbzConverter:
    return CbzConverter



# because no natsort so you would get
# 1, 10, 2, 3, 4, 41, 42, 5
import imghdr
import json
import multiprocessing as mp
import os
import urllib.parse
from pathlib import Path
from typing import Iterable, Sequence

import requests

from .comic import BaseChapter, BaseMetadata, Comic

NUM_LEFT_PAD_DIGITS = 5
FILE_PADDING = f"0{NUM_LEFT_PAD_DIGITS}"
MD_METADATA_FILE = "md-metadata.json"


def async_download_image(
    data: tuple[str, Path | str, str | None, dict[str, str] | None]
) -> None:
    url, dest_folder, filename, headers = data
    dest_folder = Path(dest_folder)

    name = filename or url.split("/")[-1]
    dest_file = dest_folder / name

    res = requests.get(url, headers=headers)
    res.raise_for_status()
    with open(dest_file, "wb") as file:
        file.write(res.content)

    # if the file extension is lying
    # rename it so epubcheck doesn't yell at us
    ext = imghdr.what(dest_file)
    if ext is not None:
        dest_file.rename(dest_file.with_suffix(f".{ext}"))


def download_images(
    urls: Sequence[str],
    dest_folder: Path | str,
    *,
    filestems: Sequence[str] | None = None,
    headers: dict[str, str] | None = None,
    threads: int = 1,
) -> Iterable[None]:
    """
    Download one or multiple URLs to a destination folder.
    Raises ValueError if the folder does not exist.
    :param `urls`: A list of URLs to download.
    :param `dest_folder`: The path to download files into.
    :param `filestems`: Specify the name of each downloaded file instead of the default.
    :param `headers`: Request headers
    :param `threads`: The number of processes to open
    """
    dest_folder = Path(dest_folder)

    # attempt to create
    dest_folder.mkdir(exist_ok=True)

    # args to async_download
    map_pool: list[tuple[str, Path | str, str | None, dict[str, str] | None]] = []

    for url, stem in zip(urls, filestems, strict=True):
        _, ext = os.path.splitext(urllib.parse.urlparse(url).path)
        map_pool.append((url, dest_folder, f"{stem}{ext}", headers))

    with mp.Pool(threads) as pool:
        yield from pool.imap_unordered(async_download_image, map_pool)


def read_comic(path: Path | str) -> Comic:
    path = Path(path)
    json_path = path / MD_METADATA_FILE

    with open(json_path, "r", encoding="utf-8") as file:
        data = json.load(file)

    return Comic(
        BaseMetadata(**data["metadata"]),
        [BaseChapter(**c) for c in data["chapters"]],
    )


def save_comic(comic: Comic, path: Path | str) -> None:
    path = Path(path)
    path.mkdir(exist_ok=True)

    json_path = path / MD_METADATA_FILE

    with open(json_path, "w", encoding="utf-8") as file:
        json.dump(comic.asdict(), file)



from pathlib import Path
from typing import Iterable

from . import iohandler, sources
from .comic import Comic
from .converter import ConvertFormats, get_converter


def query(url: str) -> Comic:
    """
    Attempt to query for a comic given a URL.
    :param `url`: An internet URL to search for
    """
    adapter = sources.get_class_for(url)(url)
    return Comic(adapter.metadata, adapter.chapters)


def read(path: Path | str) -> Comic:
    """
    Load a mandown-created comic from the file system.
    :param `path`: A folder where mandown has created a comic
    """
    return iohandler.read_comic(path)


def convert_progress(
    comic: Comic,
    folder_path: Path | str,
    convert_to: ConvertFormats,
    dest_folder: Path | str | None = None,
) -> Iterable:
    # default to working directory
    dest_folder = dest_folder or Path(".").resolve()

    # obviously pylint is wrong because this is 100% callable
    converter = get_converter(convert_to)(comic)  # pylint: disable=not-callable
    yield from converter.create_file_progress(folder_path, dest_folder)


def convert(
    comic: Comic,
    folder_path: Path | str,
    convert_to: ConvertFormats,
    dest_folder: Path | str | None = None,
) -> None:
    for _ in convert_progress(comic, folder_path, convert_to, dest_folder):
        pass


def process_progress(comic: Comic) -> Iterable:
    pass


def process(comic: Comic) -> None:
    for _ in process_progress(comic):
        pass


def download_progress(
    comic: Comic,
    path: Path | str,
    *,
    start: int | None = None,
    end: int | None = None,
    threads: int = 2,
    only_download_missing: bool = True,
) -> Iterable:
    path = Path(path)

    # create dir
    try:
        full_path = path / comic.metadata.title
        full_path.mkdir(exist_ok=True)
    except IOError:
        # invalid filename
        full_path = path / comic.metadata.title_slug
        full_path.mkdir(exist_ok=True)

    # save metadata json
    comic.save_metadata(path)

    # for each chapter
    for chap in comic.chapters[start:end]:
        image_urls = comic.get_chapter_image_urls(chap)
        # expect that they're named by numbers only
        skip_images: set[int] = set()
        if only_download_missing:
            for f in path.iterdir():
                if f.stem == f.stem.rjust(iohandler.NUM_LEFT_PAD_DIGITS, "0"):
                    try:
                        skip_images.add(int(f.stem))
                    except ValueError:
                        # expected if not an image file
                        pass

        if not image_urls or len(skip_images) == len(image_urls):
            # skip processing if there's nothing to download
            continue

        # name them 00001.png, 00002.png, etc
        # skipping ones that already exist
        processed_image_urls, filestems = zip(
            *(
                (link, str(i).rjust(iohandler.NUM_LEFT_PAD_DIGITS, "0"))
                for i, link in enumerate(image_urls, start=1)
                if i not in skip_images
            )
        )

        chapter_path = full_path / chap.slug

        yield from iohandler.download_images(
            processed_image_urls,
            chapter_path,
            headers=comic.source.headers,
            filestems=filestems,
            threads=threads,
        )


def download(
    comic: Comic,
    path: Path | str,
    *,
    start: int = 0,
    end: int = -1,
    threads: int = 2,
    only_download_missing: bool = True,
) -> None:
    for _ in download_progress(
        comic,
        path,
        start=start,
        end=end,
        threads=threads,
        only_download_missing=only_download_missing,
    ):
        pass



import dataclasses
from dataclasses import dataclass
from enum import Enum
from typing import Literal

from . import cbz, epub, pdf
from .base_converter import BaseConverter

CBZ = "cbz"
EPUB = "epub"
PDF = "pdf"


class ConvertFormats(str, Enum):
    # for typing purposes
    CBZ = "cbz"
    EPUB = "epub"
    PDF = "pdf"


@dataclass(kw_only=True)
class ConvertOptions:
    page_progression: Literal["ltr"] | Literal["rtl"] = "ltr"


def get_converter(convert_to: ConvertFormats) -> type[BaseConverter]:
    match convert_to:
        case ConvertFormats.CBZ:
            return cbz.CbzConverter
        case ConvertFormats.EPUB:
            return epub.EpubConverter
        case ConvertFormats.PDF:
            return pdf.PdfConverter
